SUBROUTINE compute_coupling_matrix(nlmax,M_l,Ftp,ipiv)
!
! This subroutine calculates an LU decomposition of the coupling
! matrix. Ref: Hivon et al., ApJ, 567, 2 (2002)
!
! = Inputs =
! 1. nlmax (maximum multipole) [integer]
! 2. M_l(0:lmax) [power spectrum of a mask; double precision]
!
! = Outputs =
! 1. Ftp(0:lmax-2,0:lmax-2) [double precision]
! 2. ipiv(1:lmax-1) [integer]
!
! June 27, 2010: E.Komatsu
!
! Calculation of the couplong matrix is modified to use symmetry of the square of Wigner 3j
! July 26, 2018: R. Makiya
!
  implicit none
  integer,intent(IN) :: nlmax
  double precision, intent(INOUT) :: M_l(0:nlmax)
  double precision, intent(OUT) :: Ftp(nlmax-1,nlmax-1)
  integer, intent(OUT) :: ipiv(nlmax-1)
  integer :: info,l,l2,l3,lmin,lmax
  double precision :: el1min,el1max
  double precision :: thrcof(2*nlmax+2),w3j(0:nlmax)
  double precision, parameter :: pi= 3.14159265358979323846264d0
  do l=0,nlmax
     M_l(l)=(2d0*dble(l)+1d0)*M_l(l)
  enddo
! compute TT-TT coupling
  Ftp=0d0
  do l2=2,nlmax
     do l3=l2,nlmax
        CALL DRC3JJ(dble(l2),dble(l3),0.d0,0.d0,EL1MIN,EL1MAX, &
             THRCOF,2*nlmax+2,info)
        lmin=INT(EL1MIN)
        lmax=MIN(nlmax,INT(EL1MAX))
        w3j(lmin:lmax)=thrcof(1:lmax-lmin+1)**2d0
        Ftp(l2-1,l3-1)=SUM(M_l(lmin:lmax)*w3j(lmin:lmax)) 
        if (l3 /= l2) then
          Ftp(l3-1,l2-1) = Ftp(l2-1,l3-1)
        endif
     enddo
  enddo
  do l3=2,nlmax
    Ftp(1:nlmax-1,l3-1) = (2d0*l3+1d0)/(4d0*pi)*Ftp(1:nlmax-1,l3-1)
  end do
! LU decompose the coupling matrix
  CALL DGETRF(nlmax-1,nlmax-1,Ftp,nlmax-1,ipiv,info)
  if(info.ne.0)then
     print*,' error DGETRF: info =',info
     stop
  endif
  RETURN
END SUBROUTINE compute_coupling_matrix
!-------------------------------------------------------------------------
SUBROUTINE COMPUTE_CL(nlmax,cl,Ftp,ipiv)
!
! This subroutine solves the following linear equation:
!
!    Ftp * C_l = \tilde{C}_l
!
! for C_l.
!
! The input Ftp is an LU decomposition of the original matrix 
! computed by DGETRF.
!
! = Inputs =
! 1. nlmax [integer]
! 2. cl(0:nlmax) [double precision] -> also output
! 3. Ftp(0:nlmax-2,0:nlmax-2) [double precision]
! 4. ipiv(1:nlmax-1) [integer]
!
! December 18, 2003: E.Komatsu
!
  IMPLICIT NONE
  integer, INTENT(IN) :: nlmax
  double precision, INTENT(IN) :: Ftp(nlmax-1,nlmax-1)
  integer, INTENT(IN) :: ipiv(nlmax-1)
  double precision, INTENT(INOUT) :: cl(0:nlmax)
  double precision, ALLOCATABLE :: cls(:)
  integer :: info
  ALLOCATE(cls(nlmax-1))
  cls(1:nlmax-1)=cl(2:nlmax)
  CALL DGETRS('N',nlmax-1,1,Ftp,nlmax-1,ipiv,cls,nlmax-1,info)
  if(info.ne.0)then
     print*,' error at DGETRS in COMPUTE_CL : info = ',info
  endif
  cl(2:nlmax)=cls(1:nlmax-1)
  DEALLOCATE(cls)
  RETURN
END SUBROUTINE COMPUTE_CL
